from .._process import Process
from dask.array import Array
import dask.array as da
from skimage.segmentation import expand_labels

class LabelDilation(Process):
    def __init__(self, distance=1):
        super().__init__("Perform label dilation")
        self.distance = distance

    def process(self, label_image: Array) -> Array:
        assert isinstance(label_image, Array), f"Expected label_image to be of type Array, got {type(label_image)}"
        assert label_image.ndim == 2, f"Expected label_image to be 2D, got {label_image.ndim}D"
        assert label_image.dtype in (int, 'int32', 'int64'), f"Expected label_image to be of type int, got {label_image.dtype}"
        distance = self.distance
        overlap = 2 * distance + 1
        label_image = da.map_overlap(lambda x: expand_labels(x, distance=distance), label_image, depth=overlap, boundary='none', dtype=label_image.dtype)
        return label_image


from .._process import Process
import dask.array as da
from dask.array import Array
from skimage.color import rgb2hsv

class RGBToHSV(Process):
    def __init__(self) -> None:
        super().__init__("Convert an RGB image to an HSV image.")

    def process(self, image: Array) -> Array:
        assert isinstance(image, da.Array), f"Expected image to be dask.array.Array, got {type(image)}"
        assert image.ndim == 3, f"Expected image to be 3D, got {image.ndim}D"
        assert image.dtype == "uint8", f"Expected image to be uint8, got {image.dtype}"
        image = da.map_blocks(rgb2hsv, image, dtype="float64")
        return image


from ..process._processor import Processor
from abc import ABC, abstractmethod
from dask.array import Array

class QCStep(ABC):
    def __init__(self, name, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        self.name = name
        preprocessor = preprocessor or self._default_preprocessor()
        postprocessor = postprocessor or self._default_postprocessor()
        self.preprocessor = preprocessor
        self.postprocessor = postprocessor

    @abstractmethod
    def fit(self, image: Array) -> None:
        pass

    @abstractmethod
    def run(self, image: Array) -> Array:
        pass

    @abstractmethod
    def fit_and_run(self, image: Array) -> Array:
        pass

    @abstractmethod
    def _default_preprocessor(self) -> Processor:
        return Processor()

    @abstractmethod
    def _default_postprocessor(self) -> Processor:
        return Processor()


from .._process import Process
import dask.array as da
from dask.array import Array

class ABContrast2D(Process):
    def __init__(self, a: float = 1, b: float = 0):
        super().__init__("Apply a linear contrast transformation to a 2D image.")
        assert isinstance(a, float), "a must be a float"
        assert isinstance(b, float), "b must be a float"
        self.a = a
        self.b = b

    def process(self, image: Array) -> Array:
        image = image * self.a + self.b
        image = da.clip(image, 0, 255)
        image = image.astype('uint8')
        return image

class GammaContrast2D(Process):
    def __init__(self, gamma: float = 1):
        super().__init__("Apply a gamma contrast transformation to a 2D image.")
        assert isinstance(gamma, float), "gamma must be a float"
        self.gamma = gamma

    def process(self, image: Array) -> Array:
        assert isinstance(image, da.Array), "image must be a dask array"
        assert image.ndim == 2, "image must be a 2D array"
        assert image.dtype == 'uint8', "image must be a uint8 array"
        image = image / 255
        image = image ** self.gamma
        image = image * 255
        image = image.astype('uint8')
        return image


from .._qc_step import QCStep
from ...process._processor import Processor

class PenDetector(QCStep):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        super().__init__(name, preprocessor, postprocessor)


from ._process import Process
from dask.array import Array

class Processor():
    def __init__(self):
        self.processes = []

    def add_process(self, process: Process) -> None:
        assert isinstance(process, Process), f"Expected process to be Process, got {type(process)}"
        self.processes.append(process)

    def process(self, image: Array) -> Array:
        for process in self.processes:
            image = process.process(image)
        return image

    def __str__(self) -> str:
        text = f"Processor with {len(self.processes)} processes:"
        for i, process in enumerate(self.processes):
            text += f"\n\t({i+1}) {process.name}"
        return text

    def __repr__(self) -> str:
        return self.__str__()


from .._segmenters import PrimarySegmenter
from ...process._processor import Processor
from ...process.colour import RGBToOD
import dask.array as da
from dask.array import Array
from skimage.filters import threshold_otsu

class GlobalOtsuTissueSegmenter(PrimarySegmenter):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None, invert_foreground: bool = False) -> None:
        super().__init__("Global Otsu Thresholder Tissue Segmenter", preprocessor, postprocessor)
        assert isinstance(invert_foreground, bool), "Invert foreground must be a boolean"
        self.invert_foreground = invert_foreground

    def fit(self, image: Array) -> None:
        image = self.preprocessor.process(image)
        assert isinstance(image, da.Array), "Image must be a dask array"
        assert image.ndim == 2, "Image must be 2D"
        assert image.dtype == "uint8", "Image must be uint8"
        counts, _ = da.histogram(image, bins=256, range=[0, 256])
        self.threshold = threshold_otsu(hist=counts.compute(), nbins=256)
        self.fitted = True

    def segment(self, image: Array) -> Array:
        assert self.fitted, "Segmenter must be fitted before segmenting"
        image = self.preprocessor.process(image)
        assert isinstance(image, da.Array), "Image must be a dask array"
        assert image.ndim == 2, "Image must be 2D"
        assert image.dtype == "uint8", "Image must be uint8"
        threshold = self.threshold
        if self.invert_foreground:
            image = da.less(image, threshold)
        else:
            image = da.greater_equal(image, threshold)
        return self.postprocessor.process(image)

    def fit_and_segment(self, image: Array) -> Array:
        self.fit(image)
        return self.segment(image)

    def _default_preprocessor(self) -> Processor:
        return Processor().add_process(RGBToOD())

    def _default_postprocessor(self) -> Processor:
        return Processor()


from abc import ABC, abstractmethod
from dask.array import Array

class Process(ABC):
    def __init__(self, name: str) -> None:
        assert isinstance(name, str), f"Expected name to be str, got {type(name)}"
        self.name = name

    @abstractmethod
    def process(self, image: Array) -> Array:
        pass

    def __str__(self) -> str:
        return f"Process: {self.name}"

    def __repr__(self) -> str:
        return self.__str__()


from .._segmenters import PrimarySegmenter
from ...process._processor import Processor
from ...process.colour._hsv_transform import RGBToHSV
import dask.array as da
from dask.array import Array

class RulesBasedTissueSegmenter(PrimarySegmenter):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        super().__init__("Rules Based Tissue Segmenter", preprocessor, postprocessor)
        self.fitted = True

    def fit(self, image: Array) -> None:
        pass

    def segment(self, image: Array) -> Array:
        image = self.preprocessor.process(image)
        image = RGBToHSV().process(image)
        saturation = image[:, :, 1]
        saturation = da.less_equal(saturation, 0.1)
        value = image[:, :, 2]
        value = da.greater_equal(value, 0.1)
        blank_mask = da.logical_and(saturation, value)
        hue = image[:, :, 0]
        hue1 = da.less(hue, 0.7)
        hue2 = da.greater(hue, 0.4)
        hue = da.logical_and(hue1, hue2)
        saturation = image[:, :, 1]
        saturation = da.greater(saturation, 0.1)
        hue_saturation = da.logical_and(hue, saturation)
        value = image[:, :, 2]
        value = da.less(value, 0.1)
        blue_mask = da.logical_or(hue_saturation, value)
        mask = da.logical_and(blank_mask, blue_mask)
        mask = da.logical_not(mask)
        mask = self.postprocessor.process(mask)
        return mask

    def fit_and_segment(self, image: Array) -> Array:
        return self.segment(image)

    def _default_preprocessor(self) -> Processor:
        return Processor()

    def _default_postprocessor(self) -> Processor:
        return Processor()



from .._process import Process
from dask.array import Array
from dask_image.ndmorph import binary_closing, binary_dilation, binary_erosion, binary_opening

class BinaryClosing(Process):
    def __init__(
        self,
        structure = None,
        iterations=1,
        origin=0,
        mask=None,
        border_value=0,
        brute_force=False
    ) -> None:
        super().__init__("Perform binary closing on a 2D array.")
        self.structure = structure
        self.iterations = iterations
        self.origin = origin
        self.mask = mask
        self.border_value = border_value
        self.brute_force = brute_force

    def process(self, image: Array) -> Array:
        assert isinstance(image, Array), "image must be a dask array"
        assert image.ndim == 2, "image must be a 2D array"
        return binary_closing(
            image,
            structure=self.structure,
            iterations=self.iterations,
            origin=self.origin,
            mask=self.mask,
            border_value=self.border_value,
            brute_force=self.brute_force
        )

class BinaryDilation(Process):
    def __init__(
        self,
        structure = None,
        iterations=1,
        origin=0,
        mask=None,
        border_value=0,
        brute_force=False
    ) -> None:
        super().__init__("Perform binary dilation on a 2D array.")
        self.structure = structure
        self.iterations = iterations
        self.origin = origin
        self.mask = mask
        self.border_value = border_value
        self.brute_force = brute_force

    def process(self, image: Array) -> Array:
        assert isinstance(image, Array), "image must be a dask array"
        assert image.ndim == 2, "image must be a 2D array"
        return binary_dilation(
            image,
            structure=self.structure,
            iterations=self.iterations,
            origin=self.origin,
            mask=self.mask,
            border_value=self.border_value,
            brute_force=self.brute_force
        )

class BinaryErosion(Process):
    def __init__(
        self,
        structure = None,
        iterations=1,
        origin=0,
        mask=None,
        border_value=0,
        brute_force=False
    ) -> None:
        super().__init__("Perform binary erosion on a 2D array.")
        self.structure = structure
        self.iterations = iterations
        self.origin = origin
        self.mask = mask
        self.border_value = border_value
        self.brute_force = brute_force

    def process(self, image: Array) -> Array:
        assert isinstance(image, Array), "image must be a dask array"
        assert image.ndim == 2, "image must be a 2D array"
        return binary_erosion(
            image,
            structure=self.structure,
            iterations=self.iterations,
            origin=self.origin,
            mask=self.mask,
            border_value=self.border_value,
            brute_force=self.brute_force
        )

class BinaryOpening(Process):
    def __init__(
        self,
        structure = None,
        iterations=1,
        origin=0,
        mask=None,
        border_value=0,
        brute_force=False
    ) -> None:
        super().__init__("Perform binary opening on a 2D array.")
        self.structure = structure
        self.iterations = iterations
        self.origin = origin
        self.mask = mask
        self.border_value = border_value
        self.brute_force = brute_force

    def process(self, image: Array) -> Array:
        assert isinstance(image, Array), "image must be a dask array"
        assert image.ndim == 2, "image must be a 2D array"
        return binary_opening(
            image,
            structure=self.structure,
            iterations=self.iterations,
            origin=self.origin,
            mask=self.mask,
            border_value=self.border_value,
            brute_force=self.brute_force
        )


import dask.array as da
import numpy as np
from scematk.process.colour import RGBToGrey
from skimage import img_as_ubyte
from skimage.color import rgb2gray

def test_RGBToGrey():
    img = np.random.randint(0, 255, size=(100, 100, 3), dtype="uint8")
    ref = img_as_ubyte(rgb2gray(img))
    img = da.from_array(img, chunks=(50, 50, 3))
    test = RGBToGrey().process(img).compute()
    equality = ref == test
    assert equality.all()